/*
 * config.c - functions for reading and setting the configuration
 *
 * Copyright (c) 2003
 *    Felix Opatz <felix@zotteljedi.de>.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer as
 *    the first lines of this file unmodified.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY FELIX OPATZ ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL FELIX OPATZ BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


/* a short introduction to this file:
 *
 * exported functions are config_load_file, config_print (when
 * DEBUG is defined), config_set_defaults and config_set_ids.
 *
 * config_set_defaults simply sets the defaults to a fresh
 * allocated struct server_config{}.
 *
 * config_print is called at the end of config_load_file and
 * prints out what is actually loaded.
 *
 * config_set_ids is called by config_load_file when the uid
 * or gid has to be changed. Note: it actually NOT sets the
 * IDs (perhaps the name is a little bit misleading), it only
 * resolved usernames and groupnames to the IDs (=> sets the
 * ID-fields of the struct, set_ids...)
 *
 * config_load_file is the most interesting function, because
 * it loads and parses the configuration file. It calls a
 * local function parse() that makes a key/value pair from
 * a given line in the config-file. There is also a local
 * function is_comment_or_empty(), I think the name says
 * everything you want to know. :-)
 *
 * This file (and especially config_load_file) is the place
 * where the configuration of new options should be implemented
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <pwd.h>
#include <grp.h>

#ifdef DEBUG
#include <unistd.h>
#endif

#include "zjhttpd.h"

/* only needed in this file */
enum CONFIG_ERRORS {CONFIG_SUCCESS, CONFIG_NOVALUE,
		CONFIG_EMPTY, CONFIG_NOQUOTE, CONFIG_QUOTE};

/* This function sets the default values */
int config_set_defaults(struct server_config *conf)
{
	conf->www_dir = DEFAULT_WWW_DIR;
	conf->index = DEFAULT_INDEX;
	conf->default_ct = DEFAULT_CONTENT_TYPE;

	return 0;
}

/* This function takes a line from the configuration file and splits
 * it into a key- and value-pair. If a problem occures, the function
 * returns a non-zero value and the user will be scolded.
 * NOTE: no buffer overflows in the while-loops because key and
 * value are as big as line is.
 */
static int parse(const char *line, char *key, char *value)
{
	/* is the value-string quoted? */
	int quoted = 0;
	
	/* skip leading whitespaces */
	while (*line && isspace(*line))
		line++;

	/* read the key which is a single word, so stop when
	 * whitespaces or the = occurs
	 */
	while (*line && !isspace(*line) && *line != '=')
		*(key++) = *(line++);

	/* null-terminate the key-string */
	*key = '\0';

	/* skip whitespaces between key and = */
	while (*line && isspace(*line))
		line++;

	/* huh, no value? */
	if (!*line || *line != '=')
		return CONFIG_NOVALUE;

	/* skip whitespaces and the = */
	while (*line && (isspace(*line) || *line == '='))
		line++;

	/* empty value but no quotes -> warning */
	if (!*line)
		return CONFIG_EMPTY;

	/* if the value-string is quoted, skip the " */
	if (*line && *line == '"')
	{
		quoted = 1;
		line++;
	}

	/* read the value until end of line, newline, carriage return
	 * or a terminating "
	 */
	while (*line && *line != '"' && *line != '\r' && *line != '\n')
		*(value++) = *(line++);

	/* null-terminate the value-string */
	*value = '\0';

	/* quoted value but no end of quote? */
	if (quoted && *line != '"')
		return CONFIG_NOQUOTE;

	/* an ending quote without beginning? */
	if (!quoted && *line == '"')
		return CONFIG_QUOTE;

	return CONFIG_SUCCESS;
}

/* this function determines whether a line is empty or a comment */
static int is_comment_or_empty(const char *line)
{
	while (*line && isspace(*line))
		line++;

	return (*line == '#') || (!*line);
}

/* this function loads the configuration from a file */
int config_load_file(struct server_config *conf, const char *filename,
		int quite)
{
	FILE *file;
	char line[BUF_SIZ], key[BUF_SIZ], value[BUF_SIZ];
	unsigned long line_number = 0;
	char *user = NULL, *group = NULL;

	file = fopen(filename, "r");
	if (!file)
	{
		/* quite mode if called for the default config file */
		if (!quite)
			fprintf(stderr, "error opening %s: %s\n",
					filename, strerror(errno));
		return 1;
	}

	/* read until end of line */
	while (fgets(line, sizeof(line), file))
	{
		line_number++;

		/* skip comments and empty lines */
		if (is_comment_or_empty(line))
			continue;
		
		/* parse line and test for errors */
		switch(parse(line, key, value))
		{
			case CONFIG_SUCCESS:
				break;

			/* no value specified: error */
			case CONFIG_NOVALUE:
				fprintf(stderr, "%s:%lu: no value found\n",
								filename, line_number);
				return 2;

			/* quoted string not finished: error */
			case CONFIG_NOQUOTE:
				fprintf(stderr, "%s:%lu: no end of quote found\n",
								filename, line_number);
				return 3;

			/* ending quote without starting a quoted string: error */
			case CONFIG_QUOTE:
				fprintf(stderr, "%s:%lu: end of quote found, nut no "
								"start\n", filename, line_number);
				return 4;

			/* empty value not quoted: warning */
			case CONFIG_EMPTY:
				fprintf(stderr, "%s:%lu: suggest quotes around empty "
								"values\n", filename, line_number);
				break;
		}

#ifdef DEBUG
		printf("config: key = <%s> value = <%s>\n", key, value);
#endif

		if (!cmp(key, "user"))
		{
			user = strdup(value);
			continue;
		}

		if (!cmp(key, "group"))
		{
			group = strdup(value);
			continue;
		}

		if (!cmp(key, "www_dir"))
		{
			conf->www_dir = strdup(value);
			continue;
		}

		if (!cmp(key, "index"))
		{
			conf->index = strdup(value);
			continue;
		}

		if (!cmp(key, "error_file_path"))
		{
			conf->error_file_path = strdup(value);
			continue;
		}

		if (!cmp(key, "dotfiles"))
		{
			if (!cmp(value, "yes"))
				conf->dotfiles = 1;
			continue;
		}

		if (!cmp(key, "default_content_type"))
		{
			conf->default_ct = strdup(value);
			continue;
		}

		if (!cmp(key, "extended_logging"))
		{
			if (!cmp(value, "yes"))
				conf->extlog = 1;
			continue;
		}

		if (!cmp(key, "dirlisting"))
		{
			if (!cmp(value, "yes"))
				conf->dirlisting = 1;
			continue;
		}

		fprintf(stderr, "%s:%lu unknown option \"%s\"\n",
				filename, line_number, key);
		return 5;
	}

	if ((user || group) && config_set_ids(conf, user, group))
		return 6;

	return 0;
}

/* set the user ID and group ID, resolve if needed */
int config_set_ids(struct server_config *conf, const char *user,
	const char *group)
{
	unsigned long val;
	char *end;
	struct passwd *pwd;
	struct group *grp;

	/* should the user ID be changed? */
	if (user)
	{
		/* able to convert into a number? then it is ok */
		val = strtoul(user, &end, 0);
		if (*end)
		{
			/* resolve username into uid */
			pwd = getpwnam(user);
			if (!pwd)
			{
				fprintf(stderr, "couldn't find user \"%s\"\n", user);
				return 1;
			}
			conf->uid = pwd->pw_uid;
		}
		else
			conf->uid = (uid_t) val;
	}

	/* should the group ID be changed? */
	if (group)
	{
		/* numeric group ID? */
		val = strtoul(group, &end, 0);
		if (*end)
		{
			/* if not, resolve it */
			grp = getgrnam(group);
			if (!grp)
			{
				fprintf(stderr, "couldn't find group \"%s\"\n", group);
				return 2;
			}
			conf->gid = grp->gr_gid;
		}
		else
			conf->gid = (gid_t) val;
	}

	return 0;
}

#ifdef DEBUG
/* print out the configuration */
void config_print(const struct server_config *conf)
{
	printf("configuration is:\n");

	printf("www_dir = <%s>\n", conf->www_dir);

	printf("index = <%s>\n", conf->index);

	if (conf->uid)
		printf("uid = %i\n", conf->uid);
	else
		printf("uid won't be changed, but is currently %u\n", getuid());

	if (conf->gid)
		printf("gid = %i\n", conf->gid);
	else
		printf("gid won't be changed, but is currently %u\n", getgid());
}
#endif
