/*
 * response.c - create a proper response, DO THE WORK! :-)
 * 
 * Copyright (c) 2003
 *    Felix Opatz <felix@zotteljedi.de>.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer as
 *    the first lines of this file unmodified.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY FELIX OPATZ ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL FELIX OPATZ BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/param.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <time.h>
#include <unistd.h>

#include "zjhttpd.h"

/* a short introduction to this file:
 *
 * All the responses are created here (or in the specialized functions).
 * Entry point is send_response (called by main_loop from request.c).
 *
 * We need to copy some things from http_request to http_response. I
 * tried to minimize the parameters to pass down to the functions by
 * introducing a new struct with the values the response-functions need.
 * Well, it didn't work as good as I hoped, but it still is okay.
 *
 * First the file is stat'ed (stat_file). We read the symlinks in
 * order to implement the 301 and 309 responses for moved files. If
 * a symlink is local, we send 309, if it is external (starting with
 * things like http:// or ftp://) we send 301 responses. If it is
 * a directory, deal_directory() is called (which fills a few values
 * of http_response that would have been filled in by stat_file, but
 * this function is left with return STAT_IS_DIRECTORY).
 *
 * If a user requests a directory without a trailing /, a moved
 * temporarily message (301) will be sent.
 *
 * deal_directory will determine whether a index.html is existing
 * (if so, it is sent), or if not whether a 403 should be sent,
 * or a directory listing.
 *
 * Note that deal_directory calls a function from misc.c that will
 * send the directory listing (haha, that's why this file is so short,
 * it only delegates the work ;-)).
 *
 * If-Modified-Since and If-Unmodified-Since is done here (entry point
 * function) and after that a switch will determine which function
 * to call to generate the specialized response.
 *
 * Currently (0.1.0) only GET and HEAD are implemented, but later
 * POST and maybe other (PUT? Uargs) may be implemented by adding
 * code (for calling specialized code!) there.
 *
 * Note the strange bytes_sent = content_sent stuff at the end, if
 * you don't understand this, read the comments in misc.c concerning
 * the logfile-function.
 *
 * Note: I hope that all the parsing stuff is done before this point,
 * to that the following functions (e.g. http_post from post.c in later
 * verisons) will only have to work with the values. BUT: if a request
 * is sent, that has a body (and a Content-Length), it is left in the
 * buffer and MUST be retrieved by fread (no low level functions!).
 * If you don't get the stuff from stdin, the rest of the Keep-Alive
 * session will not work!
 *
 * Currently the server will produce problems if a clients sends a
 * body, but I didn't want to write a function that reads the stuff
 * and throws it away, because a) this server is in an early state
 * and it is ok if it doesn't work with any input, and b) I want to
 * do it once right instead of implementing it twice. Are GET and
 * HEAD with a body allowed? I don't think so, so it even won't be
 * an offence against RFC 2616 / HTTP/1.1.
 */

/* stat(2) the file and check if failed why */
static int stat_file(const struct server_config *conf,
		const struct http_request *req, struct http_response *res)
{
	char full_path[BUF_SIZ * 2];	/* just to be sure */
	char resolved_path[MAXPATHLEN + 1], *p;
	struct stat stbuf;
	int ret;
	time_t now;

	/* first time: put the paths together */
	snprintf(full_path, sizeof(full_path), "%s/%s", conf->www_dir, req->path);

	/* does the file exist? Permissions? */
	if (lstat(full_path, &stbuf) == -1)
	{
		if (errno == EACCES)
			return HTTP_FORBIDDEN;
		return HTTP_NOT_FOUND;
	}

	/* is it a symlink? */
	if ((stbuf.st_mode & S_IFLNK) == S_IFLNK)
	{
		if((ret = readlink(full_path, resolved_path,
			sizeof(resolved_path))) == -1)
			return HTTP_NOT_FOUND;
		resolved_path[ret] = '\0';

		/* external links: http://, ftp://, ... */
		if (strstr(resolved_path, "://"))
		{
			res->path = strdup(resolved_path);
			return HTTP_MOVED_PERMANENTLY;
		}

		realpath(full_path, resolved_path);

		p = resolved_path + strlen(conf->www_dir);
		if (*p != '/')
			p--;

		res->path = strdup(p);

		return HTTP_MOVED_TEMPORARILY;
	}

	/* if we reach here, the file exists and is an ordinary file.
	 * Well, no, it may be a FIFO etc., but we treat it as an
	 * ordinary file to be more flexible (perhaps the user wants
	 * the server to read from a FIFO...)
	 * It also can be a directory, but we will deal with that
	 * after the realpath-expanding.
	 */

	realpath(full_path, resolved_path);

	/* hoho, GET ../../../etc/passwd, eh? ;-) */
	if (strncmp(resolved_path, conf->www_dir, strlen(conf->www_dir)))
			return HTTP_FORBIDDEN;

	/* what about dotfiles? */
	if (!conf->dotfiles)
	{
		/* we don't want to check for dotfiles in the www_dir, but
		 * we use -1 if the user names the www_dir with a trailing
		 * slash (as he should), and we only trigger on /. for
		 * dotfiles. If the user does something like /var/www////,
		 * and the //// are truncated to / by realpath, he is
		 * stupid and this function won't get the first dotfile(s)
		 * (depending on the count of /)
		 */
		for (p = resolved_path + strlen(conf->www_dir) - 1; *p; p++)
		{
			if ((*p == '/') && (*(p + 1) == '.'))
				return HTTP_FORBIDDEN;
		}
	}

	/* what about directories? */
	if ((stbuf.st_mode & S_IFDIR) == S_IFDIR)
	{
		/* is it terminated with a '/'? */
		if (req->path[strlen(req->path) - 1] != '/')
		{
			snprintf(full_path, sizeof(full_path), "%s/", req->path);
			res->path = strdup(full_path);
			return HTTP_MOVED_PERMANENTLY;
		}

		/* don't make the same stupid thing: a directory should end
		 * on a '/'!
		 */
		snprintf(full_path, sizeof(full_path), "%s/", resolved_path);
		res->path = strdup(full_path);

		return STAT_IS_DIRECTORY;
	}

	res->path = strdup(resolved_path);

	/* this part is copied to a few pages below
	 * for deal_directory. If something is broken here, also
	 * correct it below.
	 */
	res->content_length = stbuf.st_size;
	res->content_type = get_content_type(resolved_path, conf);
	res->last_modified = stbuf.st_mtime;

	/* check for future timestamps */
	if ((now = time(NULL)) < stbuf.st_mtime)
		res->last_modified = now;

	return 0;
}

/* send an apropriate moved-message, always text/plain */
static unsigned long http_moved(const struct http_response *res, int code)
{
	char buf[BUF_SIZ * 2];
	int length;
	
	length = snprintf(buf, sizeof(buf), "Your file has moved %s here: %s\r\n",
			(code == HTTP_MOVED_PERMANENTLY) ? "permanently" : "temporarily",
			res->path);
	
	printf("HTTP/%i.%i %i %s\r\n", res->http_major, res->http_minor,
			code, response_message_short(code));
	printf("Date: %s\r\n", http_datestring(time(NULL)));
	printf("Server: %s\r\n", http_serverstring());
	printf("Accept-Ranges: bytes\r\n");
	printf("Content-Type: text/plain\r\n");
	printf("Content-Length: %i\r\n", length);
	printf("Connection: %s\r\n",
			(res->keep_alive == KEEP_ALIVE_YES) ? "Keep-Alive" : "Close");
	printf("Location: %s\r\n\r\n", res->path);

	printf("%s", buf);
	
	fflush(stdout);

	return length;
}

/* free the few strdups etc. from struct http_response{}
 * In fact it is only one, but:
 * 1) I don't want to grab the structure directly to keep it blackboxed
 * 2) there may be later more things to free, just remember the plan
 *    to implement CGI later, Accept-Language, Accept-Encoding...
 */
static void free_response(const struct http_response *res)
{
	free(res->path);
}

/* deal with a requested directory */
static int deal_directory(const struct server_config *conf,
		const struct http_request *req, struct http_response *res)
{
	char buf[BUF_SIZ * 2];	/* may be a few bytes longer than BUF_SIZ */
	struct stat stbuf;
	time_t now;

	/* TODO:
	 * if index exist: send index
	 * if dirlisting on: send dir
	 * if both not: send 403
	 * adjust res.content_sent and path
	 */

	/* does it have an index file? */
	snprintf(buf, sizeof(buf), "%s/%s", res->path, conf->index);
	if (!(stat(buf, &stbuf)))
	{
		/* yes: return DIR_HAS_INDEX, adjust path */
		if ((stbuf.st_mode & S_IFREG) == S_IFREG)
		{
			free(res->path);
			res->path = strdup(buf);
			/*
			 * NOTE: here is code from above, stat_file(),
			 * because we have returned from that function a
			 * little bit to early (if it would have been a
			 * regular file). We have to fill in some things
			 * to res in order to have the GET/HEAD stuff
			 * working. If there is something broken, remember
			 * to repair it in stat_file(), too.
			 */

			res->content_length = stbuf.st_size;
			res->content_type = get_content_type(res->path, conf);
			res->last_modified = stbuf.st_mtime;

			/* check for future timestamps */
			if ((now = time(NULL)) < stbuf.st_mtime)
				res->last_modified = now;

			return DIR_HAS_INDEX;
		}
	}

	/* no? OK, are we allowed to browse directories? */
	if (!conf->dirlisting)
		/* no? Hmm, give a 403 */
		return HTTP_FORBIDDEN;

	/* yes? OK, now send the dirlisting */
	return http_send_dirlisting(req, res);
}


/* so, it's down to us to send the file, read the symlinks,
 * start the CGI stuff (not in 0.1.0) and do the rest.
 */
int send_response(const struct server_config *conf, struct http_request *req)
{
	struct http_response res;

	/* if ret is not changed, it definetly IS an Internal Server Error */
	int ret = HTTP_INTERNAL_SERVER_ERROR;

	memset(&res, 0, sizeof(res));

	/* SORRY! I know it would have been better not to adjust the
	 * HTTP-version in the request, but introduce the response
	 * earlier. It is a small design flaw, but I will keep it
	 * because it will not harm the program.
	 */
	res.http_major = req->http_major;
	res.http_minor = req->http_minor;
	res.keep_alive = req->keep_alive;
	res.method = req->method;
	res.ranges = req->ranges;

	switch ((ret = stat_file(conf, req, &res)))
	{
		/* everything OK: a normal file */
		case 0:
			break;

		/* it was a symlink, send a proper 301 or 309 */
		case HTTP_MOVED_PERMANENTLY:
		case HTTP_MOVED_TEMPORARILY:
			req->bytes_sent = http_moved(&res, ret);
			free_response(&res);
			return ret;

		/* it was a directory */
		case STAT_IS_DIRECTORY:
			/* deal with it */
			ret = deal_directory(conf, req, &res);
			switch(ret)
			{
				/* it didn't have a index and was browseable,
				 * the listing was sent without problems.
				 */
				case HTTP_OK:
					req->bytes_sent = res.content_sent;
					free_response(&res);
					return ret;

				/* the dir has an index, so GET/HEAD/whatever
				 * will work on this. res->path has been changed
				 * to res->path+index. The rest should operate
				 * on this as if it was original requested.
				 * Some addition fields of res were filled by
				 * deal_directory (see comment there).
				 */
				case DIR_HAS_INDEX:
					break;

				/* ouch, soething didn't work. We send the right
				 * error message and return with it to log it.
				 */
				default:
					req->bytes_sent = http_send_error(conf, req, ret);
					free_response(&res);
					return ret;
			}
			/* only reached by the case DIR_HAS_INDEX */
			break;

		default:
			req->bytes_sent = http_send_error(conf, req, ret);
			free_response(&res);
			return ret;
	}

	/* if we reach this point, the file is an ordinary file,
	 * does exist and we are allowed to *reach* it (not necessarily
	 * to read it). Now all the funny things like CGI or Accept-Language
	 * or Accept-Encoding may be implemented, but not in 0.1.0.
	 */

	/* Muhaha, remember If-Modified-Since and If-Not-Modified-Since?
	 * We have to implement that, too. Let's do it right here, ok?
	 * It is only interesting for GET and HEAD.
	 */

	if (((res.method == METHOD_GET) || (res.method == METHOD_HEAD))
			&& req->if_flags)
	{
		if (req->if_flags & IF_MODIFIED_SINCE)
		{
			if (res.last_modified <= req->if_modified_since)
			{
				req->bytes_sent = http_send_error(conf, req,
						HTTP_NOT_MODIFIED);
				free_response(&res);
				return HTTP_NOT_MODIFIED;
			}
		}

		if (req->if_flags & IF_UNMODIFIED_SINCE)
		{
			if (res.last_modified > req->if_unmodified_since)
			{
				req->bytes_sent = http_send_error(conf, req,
						HTTP_PRECONDITION_FAILED);
				free_response(&res);
				return HTTP_PRECONDITION_FAILED;
			}
		}
	}

	/* NOTE: it still can fail, we have an open(2) in http_get_head().
	 * We will just return HTTP_OK if everything is fine, or the
	 * error code that will be sent here - why here? Because I don't
	 * want to pass conf and req down to http_get_head().
	 */
	switch (res.method)
	{
		case METHOD_GET:
		case METHOD_HEAD:
			if ((ret = http_get_head(&res)) > HTTP_PARTIAL_CONTENT)
				req->bytes_sent = http_send_error(conf, req, ret);
			break;

		/* for later... */
		case METHOD_POST:
			break;

		case METHOD_NONE:
			ret = HTTP_INTERNAL_SERVER_ERROR;
			break;
	}

	req->bytes_sent = res.content_sent;

	free_response(&res);

	return ret;
}

